#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# __coconut_hash__ = 0xafe99d73

# Compiled with Coconut version 1.3.0 [Dead Parrot]

# Coconut Header: -------------------------------------------------------------

from __future__ import generator_stop
import sys as _coconut_sys, os.path as _coconut_os_path
_coconut_file_path = _coconut_os_path.dirname(_coconut_os_path.abspath(__file__))
_coconut_sys.path.insert(0, _coconut_file_path)
from __coconut__ import _coconut, _coconut_NamedTuple, _coconut_MatchError, _coconut_tail_call, _coconut_tco, _coconut_igetitem, _coconut_base_compose, _coconut_forward_compose, _coconut_back_compose, _coconut_forward_star_compose, _coconut_back_star_compose, _coconut_pipe, _coconut_star_pipe, _coconut_back_pipe, _coconut_back_star_pipe, _coconut_bool_and, _coconut_bool_or, _coconut_none_coalesce, _coconut_minus, _coconut_map, _coconut_partial
from __coconut__ import *
_coconut_sys.path.remove(_coconut_file_path)

# Compiled Coconut: -----------------------------------------------------------

from typing import TypeVar  # line 1
from typing import Generic  # line 1
from typing import Callable  # line 1
from typing import Union  # line 1
from typing import Any  # line 1
from typing import cast  # line 1
from typing import Optional  # line 1
from typing import Type  # line 1
from typing import Iterator  # line 1
from typing import List as TList  # line 1
from typing import Awaitable  # line 1
from functools import wraps  # line 2
from operator import eq  # line 3
import inspect  # line 4

from amino import boolean  # line 6
from amino.tc.base import F  # line 7
from amino.func import call_by_name  # line 8
from amino.func import I  # line 8
from amino.func import curried  # line 8
from amino.func import CallByName  # line 8

A = TypeVar('A')  # line 10
B = TypeVar('B')  # line 11


class Maybe(Generic[A], F[A], implicits=True):  # line 14

    __slots__ = ()  # line 16

    @_coconut_tco  # line 18
    def __new__(cls, value: 'Optional[A]') -> "'Maybe[A]'":  # line 18
        return _coconut_tail_call(Maybe.check, value)  # line 19

    @staticmethod  # line 21
    def optional(value: 'Optional[A]') -> "'Maybe[A]'":  # line 22
        return Nothing if value is None else Just(value)  # line 23

    @staticmethod  # line 25
    @_coconut_tco  # line 25
    def check(value: 'Optional[A]') -> "'Maybe[A]'":  # line 26
        return _coconut_tail_call(Maybe.optional, value)  # line 27

    @staticmethod  # line 29
    def typed(value: 'Union[A, B]', tpe: 'Type[A]') -> "'Maybe[A]'":  # line 30
        return Just(value) if isinstance(value, tpe) else Nothing  # line 31

    @staticmethod  # line 33
    def wrap(mb: "Union['Maybe[A]', None]") -> "'Maybe[A]'":  # line 34
        return mb if mb is not None and isinstance(mb, Just) else Nothing  # line 35

    @staticmethod  # line 37
    def getattr(obj: 'Any', attr: 'str') -> "'Maybe[A]'":  # line 38
        return Just(getattr(obj, attr)) if hasattr(obj, attr) else Nothing  # line 39

    @staticmethod  # line 41
    @curried  # line 41
    def iff(cond: 'bool', a: 'Union[A, Callable[[], A]]') -> "'Maybe[A]'":  # line 43
        return cast(Maybe, Just(call_by_name(a))) if cond else Nothing  # line 44

    @staticmethod  # line 46
    @curried  # line 46
    def iff_m(cond: 'bool', a: "Union['Maybe[A]', Callable[[], 'Maybe[A]']]") -> "'Maybe[A]'":  # line 48
        return cast(Maybe, call_by_name(a)) if cond else Nothing  # line 49

    @property  # line 51
    def _get(self) -> 'Union[A, None]':  # line 52
        pass  # line 53

    def cata(self, f: 'Callable[[A], B]', b: 'Union[B, Callable[[], B]]') -> 'B':  # line 55
        return (f(cast(A, self._get)) if self.is_just else call_by_name(b))  # line 56

    @_coconut_tco  # line 62
    def filter(self, f: 'Callable[[A], bool]') -> "'Maybe[A]'":  # line 62
        return _coconut_tail_call(self.flat_map, lambda a: self if f(a) else Nothing)  # line 63

    @_coconut_tco  # line 65
    def get_or_else(self, a: 'Union[A, Callable[[], A]]') -> 'A':  # line 65
        return _coconut_tail_call(self.cata, cast(Callable, I), a)  # line 66

    __or__ = get_or_else  # line 68

    @_coconut_tco  # line 70
    def get_or_raise(self, e: 'Exception') -> 'A':  # line 70
        def raise_e() -> 'None':  # line 71
            raise e  # line 72
        return _coconut_tail_call(self.cata, cast(Callable, I), raise_e)  # line 73

    @_coconut_tco  # line 75
    def get_or_fail(self, err: 'CallByName') -> 'A':  # line 75
        return _coconut_tail_call(self.get_or_raise, Exception(call_by_name(err)))  # line 76

    @_coconut_tco  # line 78
    def __contains__(self, v: 'A') -> 'boolean.Boolean':  # line 78
        return _coconut_tail_call(self.contains, v)  # line 79

    def error(self, f: 'Callable[[], Any]') -> "'Maybe[A]'":  # line 81
        self.cata(cast(Callable, I), f)  # line 82
        return self  # line 83

    def observe(self, f: 'Callable[[A], Any]') -> "'Maybe[A]'":  # line 85
        self.foreach(f)  # line 86
        return self  # line 87

    effect = observe  # line 89

    @_coconut_tco  # line 91
    def __iter__(self) -> 'Iterator':  # line 91
        return _coconut_tail_call(iter, self.to_list)  # line 92

    @property  # line 94
    @_coconut_tco  # line 94
    def is_just(self) -> 'boolean.Boolean':  # line 95
        return _coconut_tail_call(boolean.Boolean, isinstance(self, Just))  # line 96

    @property  # line 98
    def is_empty(self) -> 'boolean.Boolean':  # line 99
        return ~self.is_just  # line 100

    empty = is_empty  # line 102

    @property  # line 104
    @_coconut_tco  # line 104
    def to_list(self) -> 'TList[A]':  # line 105
        from amino.list import List  # line 106
        from amino.list import Nil  # line 106
        return _coconut_tail_call(self.cata, List, Nil)  # line 107

    @property  # line 109
    async def unsafe_await(self) -> "'Maybe[Awaitable]'":  # line 110
        if self.is_just:  # line 111
            ret = await cast(Callable[[], Awaitable], self._get)()  # line 112
            return Maybe(ret)  # line 113
        else:  # line 114
            return cast(Maybe[Awaitable], self)  # line 115

    @property  # line 117
    @_coconut_tco  # line 117
    def contains_coro(self) -> 'boolean.Boolean':  # line 118
        return _coconut_tail_call(self.exists, inspect.iscoroutine)  # line 119

    @property  # line 121
    @_coconut_tco  # line 121
    def json_repr(self) -> 'Optional[A]':  # line 122
        return _coconut_tail_call(self.cata, cast(Callable, I), lambda: None)  # line 123


class Just(Generic[A], Maybe[A]):  # line 126

    __slots__ = 'x',  # line 128

    @_coconut_tco  # line 130
    def __new__(cls, value: 'A') -> "'Maybe[A]'":  # line 130
        return _coconut_tail_call(object.__new__, cast(Type[object], cls))  # line 131

    def __init__(self, value: 'A') -> 'None':  # line 133
        self.x = value  # line 134

    @property  # line 136
    def _get(self) -> 'Optional[A]':  # line 137
        return self.x  # line 138

    @_coconut_tco  # line 140
    def __str__(self) -> 'str':  # line 140
        return _coconut_tail_call('Just({!s})'.format, self.x)  # line 141

    @_coconut_tco  # line 143
    def __repr__(self) -> 'str':  # line 143
        return _coconut_tail_call('Just({!r})'.format, self.x)  # line 144

    def __eq__(self, other: 'Any') -> 'bool':  # line 146
        return eq(self.x, other.x) if isinstance(other, Just) else False  # line 147

    @_coconut_tco  # line 149
    def __hash__(self) -> 'int':  # line 149
        return _coconut_tail_call(hash, self._get)  # line 150


class _Nothing(Generic[A], Maybe[A]):  # line 153

    __instance: 'Optional[_Nothing]' = None  # line 155

    @_coconut_tco  # line 157
    def __new__(cls: "Type['_Nothing']", *args: 'Any', **kwargs: 'Any') -> "'_Nothing[A]'":  # line 157
        if _Nothing.__instance is None:  # line 158
            _Nothing.__instance = object.__new__(cls)  # line 159
        return _coconut_tail_call(cast, _Nothing, _Nothing.__instance)  # line 160

    def __str__(self) -> 'str':  # line 162
        return 'Nothing'  # line 163

    __repr__ = __str__  # line 165

    @_coconut_tco  # line 167
    def __eq__(self, other: 'Any') -> 'bool':  # line 167
        return _coconut_tail_call(isinstance, other, _Nothing)  # line 168

    @_coconut_tco  # line 170
    def __hash__(self) -> 'int':  # line 170
        return _coconut_tail_call(hash, 'Nothing')  # line 171

Empty = _Nothing  # line 173
Nothing: Maybe = _Nothing()  # line 174


def may(f: 'Callable[..., Optional[A]]') -> 'Callable[..., Maybe[A]]':  # line 177
    @wraps(f)  # line 178
    @_coconut_tco  # line 178
    def wrapper(*a: 'Any', **kw: 'Any') -> 'Maybe[A]':  # line 179
        return _coconut_tail_call(Maybe.check, f(*a, **kw))  # line 180
    return wrapper  # line 181


def flat_may(f: 'Callable[..., Maybe[A]]') -> 'Callable[..., Maybe[A]]':  # line 184
    @wraps(f)  # line 185
    def wrapper(*a: 'Any', **kw: 'Any') -> 'Maybe[A]':  # line 186
        res = f(*a, **kw)  # line 187
        return res if isinstance(res, Maybe) else Nothing  # line 188
    return wrapper  # line 189

__all__ = ('Maybe', 'Just', 'Nothing')  # line 191



import abc
from typing import Generic, TypeVar, Callable, Tuple, cast, Type, Any

from amino.tc.base import ImplicitsMeta, F
from amino.tc.monad import Monad
from amino.tc.zip import Zip
from amino.instances.list import ListTraverse
from amino import List, Maybe, Either, Eval, IO, Left
from amino.id import Id
from amino.util.string import ToStr
from amino.tc.traverse import TraverseF, TraverseG

S = TypeVar('S')
A = TypeVar('A')
B = TypeVar('B')
ST = TypeVar('ST', bound='StateT')
G = TypeVar('G', bound=F)


class StateTMeta(ImplicitsMeta, abc.ABCMeta):

    def __new__(cls, name: str, bases: tuple, ns: dict, tpe: Type[G]=None, **kw: Any) -> Type['StateTMeta']:
        cls.monad: Monad = ...
        if tpe is not None:
            inst = super().__new__(cls, name, bases + (F,), ns, implicits=True, auto=True, **kw)
            inst.tpe = tpe  # type: ignore
            inst.monad = cast(Monad, Monad.fatal(tpe))  # type: ignore
            return inst
        else:
            return super().__new__(cls, name, bases, ns, tpe, **kw)


class StateT(Generic[G, S, A], ToStr, F[A], metaclass=StateTMeta):

    def __init__(self, run_f: F[Callable[[S], F[Tuple[S, A]]]]) -> None:
        self.run_f = run_f

    @classmethod
    def cons(self, run_f: F[Callable[[S], F[Tuple[S, A]]]]) -> 'StateT[G, S, A]':
        return self(run_f)

    @classmethod
    def apply(self, f: Callable[[S], F[Tuple[S, A]]]) -> 'StateT[G, S, A]':
        return self.cons(self.monad.pure(f))

    @classmethod
    def apply_f(self, run_f: F[Callable[[S], F[Tuple[S, A]]]]) -> 'StateT[G, S, A]':
        return self.cons(run_f)

    @classmethod
    def inspect(self, f: Callable[[S], A]) -> 'StateT[G, S, A]':
        def g(s: S) -> F[Tuple[S, A]]:
            return self.monad.pure((s, f(s)))
        return self.apply(g)

    @classmethod
    def inspect_f(self, f: Callable[[S], F[A]]) -> 'StateT[G, S, A]':
        def g(s: S) -> F[Tuple[S, A]]:
            return f(s).map(lambda a: (s, a))
        return self.apply(g)

    @classmethod
    def pure(self, a: A) -> 'StateT[G, S, A]':
        return self.apply(lambda s: self.monad.pure((s, a)))

    @classmethod
    def lift(self, fa: F[A]) -> 'StateT[G, S, A]':
        def g(s: S) -> F[Tuple[S, A]]:
            return fa.map(lambda a: (s, a))
        return self.apply(g)

    @classmethod
    def modify(self, f: Callable[[S], S]) -> 'StateT[G, S, A]':
        return self.apply(lambda s: self.monad.pure((f(s), None)))

    @classmethod
    def modify_f(self, f: Callable[[S], F[S]]) -> 'StateT[G, S, A]':
        return self.apply(lambda s: f(s).map(lambda a: (a, None)))

    @classmethod
    def set(self, s: S) -> 'StateT[G, S, A]':
        return self.modify(lambda s0: s)

    @classmethod
    def get(self) -> 'StateT[G, S, A]':
        return self.inspect(lambda a: a)

    @property
    def cls(self) -> Type[ST]:
        return cast(Type[ST], type(self))

    def run(self, s: S) -> F[Tuple[S, A]]:
        return self.run_f.flat_map(lambda f: f(s))

    def run_s(self, s: S) -> F[S]:
        return self.run(s).map(lambda a: a[0])

    def run_a(self, s: S) -> F[S]:
        return self.run(s).map(lambda a: a[1])

    def _arg_desc(self) -> List[str]:
        return List(str(self.run_f))

    def flat_map_f(self, f: Callable[[A], F[B]]) -> 'StateT[G, S, B]':
        def h(s: S, a: A) -> F[Tuple[S, B]]:
            return f(a).map(lambda b: (s, b))
        def g(fsa: F[Tuple[S, A]]) -> F[Tuple[S, B]]:
            return fsa.flat_map2(h)
        run_f1 = self.run_f.map(lambda sfsa: lambda a: g(sfsa(a)))
        return self.cls.apply_f(run_f1)

    def transform(self, f: Callable[[Tuple[S, A]], Tuple[S, B]]) -> 'StateT[G, S, B]':
        def g(fsa: F[Tuple[S, A]]) -> F[Tuple[S, B]]:
            return fsa.map2(f)
        run_f1 = self.run_f.map(lambda sfsa: lambda a: g(sfsa(a)))
        return self.cls.apply_f(run_f1)

    def modify_(self, f: Callable[[S], S]) -> 'StateT[G, S, A]':
        return self.transform(lambda s, a: (f(s), a))


def tcs(tpe: Type[G], state_tpe: Type[ST]) -> None:
    class StateMonad(Monad, tpe=state_tpe):

        def pure(self, a: A) -> StateT[G, S, A]:
            return state_tpe.pure(a)

        def flat_map(self, fa: StateT[G, S, A], f: Callable[[A], StateT[G, S, B]]) -> StateT[G, S, B]:
            def h(s: S, a: A) -> F[Tuple[S, B]]:
                return f(a).run(s)
            def g(fsa: F[Tuple[S, A]]) -> F[Tuple[S, B]]:
                return fsa.flat_map2(h)
            def i(sfsa: Callable[[S], F[Tuple[S, A]]]) -> Callable[[S], F[Tuple[S, B]]]:
                return lambda a: g(sfsa(a))
            run_f1 = fa.run_f.map(i)
            return state_tpe.apply_f(run_f1)
    class StateZip(Zip, tpe=state_tpe):

        def zip(self, fa: StateT[G, S, A], fb: StateT[G, S, A], *fs: StateT[G, S, A]) -> StateT[G, S, List[A]]:
            v = ListTraverse().sequence(cast(TraverseF[TraverseG[A]], List(fa, fb, *fs)), state_tpe)
            return cast(StateT[G, S, List[A]], v)


class MaybeState(Generic[S, A], StateT[Maybe, S, A], tpe=Maybe):
    pass

tcs(Maybe, MaybeState)


class EitherState(Generic[S, A], StateT[Either, S, A], tpe=Either):

    @classmethod
    def failed(self, err: B) -> 'StateT[Either, S, A]':
        return EitherState.lift(Left(err))

tcs(Either, EitherState)


class EvalState(Generic[S, A], StateT[Eval, S, A], tpe=Eval):
    pass

tcs(Eval, EvalState)  # type: ignore


class State(Generic[S, A], StateT[Id, S, A], tpe=Id):
    pass

IdState = State

tcs(Id, State)


class IOState(Generic[S, A], StateT[IO, S, A], tpe=IO):
    pass

tcs(IO, IOState)  # type: ignore

__all__ = ('StateT', 'MaybeState', 'EitherState', 'EvalState', 'IdState', 'State', 'IOState')



from typing import Callable, TypeVar, Type

from amino import Just, L, _, Nothing, Eval, Map
from amino.tc.monad import Monad
from amino.tc.base import ImplicitInstances, TypeClass
from amino.lazy import lazy
from amino.io import IO
from amino.util.fun import lambda_str

A = TypeVar('A')
B = TypeVar('B')


class IOInstances(ImplicitInstances):

    @lazy
    def _instances(self) -> Map[Type[TypeClass], TypeClass]:
        from amino.map import Map
        return Map({Monad: IOMonad()})


class IOMonad(Monad[IO]):

    def pure(self, a: A) -> IO[A]:
        return IO.now(a)

    def flat_map(self, fa: IO[A], f: Callable[[A], IO[B]]) -> IO[B]:
        return fa.flat_map(f)

    def map(self, fa: IO[A], f: Callable[[A], B]) -> IO[B]:
        s = Eval.later(lambda: f'map({lambda_str(f)})')
        mapper = L(f)(_) >> L(IO.now)(_)
        return fa.flat_map(mapper, Nothing, Just(s))

__all__ = ('IOInstances',)



import importlib
from typing import Optional, Any


def unsafe_import_name(modname: str, name: str) -> Optional[Any]:
    mod = importlib.import_module(modname)
    return getattr(mod, name) if hasattr(mod, name) else None


def class_path(cls: type) -> str:
    return f'{cls.__module__}.{cls.__name__}'

__all__ = ('unsafe_import_name', 'class_path')



import abc
from typing import TypeVar, Callable, Tuple, Generic

from amino.tc.functor import Functor
from amino.tc.monoidal import Monoidal

F = TypeVar('F')
A = TypeVar('A')
B = TypeVar('B')
Z = TypeVar('Z')


class Apply(Generic[F], Functor[F], Monoidal[F]):

    @abc.abstractmethod
    def ap(self, fa: F, f: F) -> F:
        ''' f should be an F[Callable[[A], B]]
        '''
        ...

    def ap2(self, fa: F, fb: F, f: Callable[[A, B], Z]) -> F:
        def unpack(tp: Tuple[A, B]):
            return f(tp[0], tp[1])
        return self.map(self.product(fa, fb), unpack)

__all__ = ('Apply',)



import os
import re
import sys
import abc
import logging
import operator
from pathlib import Path
from logging import LogRecord, DEBUG, ERROR
from typing import Callable, Any, Union, cast, Optional, TypeVar

from amino.lazy import lazy

import amino
import amino.maybe
from amino.options import EnvOption

VERBOSE = 15
TEST = 12
DEBUG1 = DDEBUG = 7
DEBUG2 = 4
logging.addLevelName(VERBOSE, 'VERBOSE')
logging.addLevelName(TEST, 'TEST')
logging.addLevelName(DEBUG1, 'DEBUG1')
logging.addLevelName(DEBUG2, 'DEBUG2')


def seq_to_str(data: Union[str, 'amino.List', Any]) -> str:
    return cast(amino.List, data).join_lines if isinstance(data, amino.List) else str(data)


class LogError(abc.ABC):

    @abc.abstractproperty
    def short(self) -> str:
        ...

    @abc.abstractproperty
    def full(self) -> str:
        ...


class LazyRecord(logging.LogRecord):

    def __init__(self, name, level, pathname, lineno, msg, args, exc_info, func=None, sinfo=None,  # type: ignore
                 extra: dict={}, **kwargs) -> None:
        super().__init__(name, level, pathname, lineno, msg, args, exc_info, func=func, sinfo=sinfo)
        self._data = msg
        self._args = args
        self._extra = extra
        self._lazy_message: Optional[str] = None

    def _cons_message(self) -> str:
        data = self._data(*self._args) if callable(self._data) else self._data
        return seq_to_str(data)

    def lazy_message(self) -> str:
        if self._lazy_message is None:
            self._lazy_message = self._cons_message()
        return self._lazy_message

    def getMessage(self) -> str:
        return (
            self._data.full
            if isinstance(self._data, LogError) else
            self.lazy_message()
            if self.levelname in ('DEBUG1', 'DEBUG2') else
            super().getMessage()
        )

    def short(self) -> str:
        try:
            short = self._extra.get('short', None)
            return self.getMessage() if short is None else short
        except Exception as e:
            return self.getMessage()


class Logger(logging.Logger):

    def test(self, message: Any, *args: Any, **kw: Any) -> None:
        if self.isEnabledFor(TEST):
            self._log(TEST, message, args, **kw)

    def verbose(self, message: Any, *args: Any, **kw: Any) -> None:
        if self.isEnabledFor(VERBOSE):
            self._log(VERBOSE, message, args, **kw)

    def debug1(self, f: Callable[..., str], *args: Any, **kw) -> None:
        if self.isEnabledFor(DEBUG1):
            self._log(DEBUG1, f, args, kw)  # type: ignore

    def debug2(self, f: Callable[..., str], *args: Any, **kw) -> None:
        if self.isEnabledFor(DEBUG2):
            self._log(DEBUG2, f, args, kw)  # type: ignore

    ddebug = debug2

    def caught_exception_error(self, when: str, exc: Exception, *a: Any, **kw: Any) -> None:
        self._caught_exception(ERROR, when, exc, *a, **kw)

    def caught_exception(self, when: str, exc: Exception, *a: Any, **kw: Any) -> None:
        self._caught_exception(DEBUG, when, exc, *a, **kw)

    def _caught_exception(self, level: int, when: str, exc: Exception, *a: Any, **kw: Any) -> None:
        headline = 'exception while {}:'.format(when)
        self.log(level, headline, exc_info=(type(exc), exc, exc.__traceback__))

    def makeRecord(self, name: str, level: int, fn: str, lno: int, msg: Any, args: Any, exc_info: Any, func: Any=None,
                   extra: Any={}, sinfo: Any=None) -> LogRecord:
        return LazyRecord(name, level, fn, lno, msg, args, exc_info, func, sinfo, extra)

    def stderr(self, message: Any) -> None:
        amino_stderr_handler.emit(self.makeRecord('stderr', ERROR, '', 0, message, (), None))


def install_logger_class() -> None:
    logging.setLoggerClass(Logger)

install_logger_class()

log: Logger = cast(Logger, logging.getLogger('amino'))
amino_root_logger = log
log.setLevel(DEBUG2)
log.propagate = False


def amino_logger(name: str) -> Logger:
    return cast(Logger, amino_root_logger.getChild(name))

_stdout_logging_initialized = False

env_log_level = EnvOption('AMINO_LOG_LEVEL')
env_xdg_runtime_dir = EnvOption('XDG_RUNTIME_DIR')
env_amino_log_dir = EnvOption('AMINO_LOG_DIR')


def init_loglevel(handler: logging.Handler, level: int=None) -> None:
    (
        amino.Maybe.check(level)
        .o(env_log_level.value)
        .o(amino.Boolean(amino.development).flat_m(VERBOSE)) %
        handler.setLevel
    )

amino_stdout_handler = logging.StreamHandler(stream=sys.stdout)
amino_stderr_handler = logging.StreamHandler()


def amino_stdout_logging(level: int=None) -> None:
    global _stdout_logging_initialized
    if not _stdout_logging_initialized:
        amino_root_logger.addHandler(amino_stdout_handler)
        init_loglevel(amino_stdout_handler, level)
        _stdout_logging_initialized = True


def log_dir() -> None:
    return env_amino_log_dir.value / Path | (
    (Path(env_xdg_runtime_dir.value | (lambda: f'/run/user/{os.getuid()}'))) / 'amino'
    )


def default_logfile() -> None:
    return log_dir() / f'log_{os.getpid()}'


_file_fmt = ('{asctime} [{levelname} @ {name}:{funcName}:{lineno}] {message}')


def amino_file_logging(logger: logging.Logger, level: int=DEBUG, logfile: Path=None, fmt: str=None
                       ) -> logging.Handler:
    file = logfile or default_logfile()
    file.parent.mkdir(exist_ok=True)
    formatter = logging.Formatter(fmt or _file_fmt, style='{')
    handler = logging.FileHandler(str(file))
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    init_loglevel(handler, level)
    return handler


def amino_root_file_logging(level: int=DEBUG, **kw: Any) -> logging.Handler:
    return amino_file_logging(amino_root_logger, level, **kw)


class Logging:

    @property
    def log(self) -> Logger:
        return self._log

    @lazy
    def _log(self) -> Logger:
        return amino_logger(self.__class__.__name__)

    def _p(self, a):
        v = self.log.verbose
        v(a)
        return a

    def _dbg(self, fmt, level=DEBUG2):
        def log(a):
            msg = fmt.format(a)
            self.log.log(level, msg)
            return a
        return log


def sub_loggers(loggers, root):
    from amino import Map, _, L
    children = loggers.keyfilter(L(re.match)('{}\.[^.]+$'.format(root), _))
    sub = (children.k / L(sub_loggers)(loggers, _)).fold_left(Map())(operator.pow)
    return Map({loggers[root]: sub})


def logger_tree(root):
    from amino import __, Map
    m = Map(logging.Logger.manager.loggerDict)
    all = m.keyfilter(__.startswith(root))
    return sub_loggers(all, 'amino')


def indent(strings: 'amino.List[str]', level: int, width: int=1) -> 'amino.List[str]':
    ws = ' ' * level * width
    return strings.map(str).map(ws.__add__)


def format_logger_tree(tree: 'amino.Map[Logger, Any]', fmt_logger: Callable[[Logger], str], level: int=0
                       ) -> 'amino.List[str]':
    from amino import _, L
    sub_f = L(format_logger_tree)(_, fmt_logger, level=level + 1)
    formatted = tree.bimap(fmt_logger, sub_f)
    return indent(formatted.flat_map2(lambda a, b: b.cons(a)), level)


def print_log_info(out: Callable[[str], None]) -> None:
    lname = lambda l: logging.getLevelName(l.getEffectiveLevel())
    hlname = lambda h: logging.getLevelName(h.level)
    def handler(h: logging.Handler) -> str:
        return '{}({})'.format(h.__class__.__name__, hlname(h))
    def logger(l: logging.Logger) -> str:
        handlers = ','.join(list(map(handler, l.handlers)))
        return '{}: {} {}'.format(l.name, lname(l), handlers)
    out(format_logger_tree(logger_tree('amino'), logger).join_lines)
    out('-------')
    out(str(env_log_level))
    out(str(amino.options.development))


A = TypeVar('A')


def with_log(f: Callable[[Logger], A], **kw: Any) -> A:
    amino_root_file_logging(**kw)
    return f(log)

__all__ = ('amino_root_logger', 'amino_stdout_logging', 'amino_file_logging', 'amino_root_file_logging',
           'print_log_info')



import re
import typing
from typing import Any, Union, Pattern

from amino import L, _, Maybe, Map, List, Either, Boolean


class Regex:

    @staticmethod
    def cons(pattern: Union[str, Pattern]) -> 'Regex':
        spec = pattern.pattern if isinstance(pattern, Pattern) else pattern
        rex = pattern if isinstance(pattern, Pattern) else re.compile(pattern)
        return Regex(spec, rex)

    def __init__(self, spec: str, rex: Union[None, Pattern]=None) -> None:
        self.spec = spec
        self.rex = re.compile(spec) if rex is None else rex

    def __getattr__(self, name: str) -> Any:
        if hasattr(self.rex, name):
            return getattr(self.rex, name)
        else:
            raise AttributeError('Regex has no attribute \'{}\''.format(name))

    def match(self, data: str, *a: Any, **kw: Any) -> Either[str, 'Match']:
        return (
            Maybe(self.rex.match(data, *a, **kw))
            .to_either('`{}` does not match `{}`'.format(data, self.spec)) /
            L(Match)(self, _, data)
        )

    def matches(self, data: str, *a: Any, **kw: Any) -> Boolean:
        return self.match(data, *a, **kw).is_right

    def search(self, data: str, *a: Any, **kw: Any) -> Either[str, 'Match']:
        return (
            Maybe(self.rex.search(data, *a, **kw))
            .to_either('`{}` does not contain `{}`'.format(data, self.spec)) /
            L(Match)(self, _, data)
        )

    def contains(self, data: str, *a: Any, **kw: Any) -> Boolean:
        return self.search(data, *a, **kw).is_right

    def __str__(self) -> str:
        return 'Regex({})'.format(self.spec)


class Match:

    def __init__(self, regex: Regex, internal: typing.Match, data: str) -> None:
        self.regex = regex
        self.internal = internal
        self.data = data

    @property
    def group_map(self) -> Map[str, str]:
        return Map(self.internal.groupdict())

    m = group_map

    def group(self, id: str) -> Either[str, str]:
        return (
            self.group_map
            .get(id)
            .to_either('no group `{}` in {}'.format(id, self))
        )

    g = group

    @property
    def groups(self) -> List[str]:
        return List.wrap(self.internal.groups())

    l = groups

    def all_groups(self, *ids: str) -> Either[str, List[str]]:
        return (self.group_map.get_all(*ids)
                .to_either('not all groups `{}` in {}'.format(ids, self)))

    @property
    def match(self) -> str:
        return self.internal.group(0)

    def __str__(self) -> str:
        return 'Match({}, {}, {})'.format(self.regex, self.data, self.group_map)

    @property
    def string(self) -> str:
        return self.internal.string

__all__ = ('Regex',)



import abc
from typing import TypeVar, Generic

from amino.tc.apply import Apply

F = TypeVar('F')
A = TypeVar('A')


class Applicative(Generic[F], Apply[F]):

    @abc.abstractmethod
    def pure(self, a: A) -> F:
        ...

__all__ = ('Applicative',)



from typing import Tuple, Generator, Any

from amino import List, Map, Lists, _, Regex, Either, Right, Left, Id, Path
from amino.state import EitherState, State
from amino.regex import Match
from amino.do import tdo
from amino.util.numeric import parse_int

Files = Map[Path, List[str]]
Entry = Map[str, Any]
excludes = List(Regex('In module imported'), Regex('__coconut__'))
error_rex = Regex('(?P<path>[^:]*):(?P<lnum>\d+)(:(?P<col>\d+))?: (error|note): (?P<error>.*)')
lnum_rex = Regex('# line (?P<lnum>\d+)')


@tdo(Either[str, Tuple[str, int, Either[str, int], str]])
def extract(match: Match) -> Generator:
    path = yield match.group('path')
    lnum = yield match.group('lnum')
    lnum_i = yield parse_int(lnum)
    col = match.group('col') // parse_int
    error = yield match.group('error')
    yield Right((Path(path), lnum_i, col, error))


def update_for(path: Path, files: Files) -> Files:
    return files if path in files else files + (path, Lists.lines(path.read_text()))


@tdo(Either[str, Entry])
def substitute(files: Files, path: Path, lnum: int, col: Either[str, int], error: str, coco_path: Path
               ) -> Either[str, Entry]:
    lines = yield files.lift(path).to_either('corrupt state')
    line = yield lines.lift(lnum - 1).to_either(f'invalid line number {lnum} for {path}')
    lnum_match = yield lnum_rex.search(line)
    coco_lnum = yield lnum_match.group('lnum')
    coco_lnum_i = yield parse_int(coco_lnum)
    col_map = col / (lambda a: Map(col=a)) | Map()
    yield Right(Map(lnum=coco_lnum_i, text=error, valid=1, maker_name='mypy') ** col_map)


@tdo(EitherState[Files, Entry])
def handle_coco(path: Path, lnum: int, col: Either[str, int], error: str, coco_path: Path) -> Generator:
    yield EitherState.modify(lambda s: update_for(path, s))
    files = yield EitherState.get()
    yield EitherState.lift(substitute(files, path, lnum, col, error, coco_path))


@tdo(EitherState[Files, Entry])
def line(l: str) -> Generator:
    r = excludes.traverse(lambda a: a.search(l).swap, Either)
    yield EitherState.lift(r)
    match = yield EitherState.lift(error_rex.match(l))
    path, lnum, col, error = yield EitherState.lift(extract(match))
    coco_path = path.with_suffix('.coco')
    yield (
        handle_coco(path, lnum, col, error, coco_path)
        if coco_path.exists() else
        EitherState.failed('not a coconut')
    )


def recover(est: EitherState[Files, Entry]) -> State[Map, Either[str, Entry]]:
    def fix(s: Map, r: Map) -> Id[Tuple[Map, Either[str, Map]]]:
        return Id((s, Right(r)))
    return State.apply(lambda s: est.run(s).map2(fix).value_or(lambda err: Id((s, Left(err)))))


def process_output(output: list) -> List[str]:
    s, result = Lists.wrap(output).map(line).traverse(recover, State).run(Map()).value
    return result.flat_map(_.to_list)

__all__ = ('process_output',)



import os
import sys
import inspect
import logging
from types import ModuleType
from typing import List, Any
from importlib.machinery import SourceFileLoader, SOURCE_SUFFIXES, FileFinder

_enable_var = 'AMINO_TYPECHECK'
_checker_var = 'AMINO_TYPECHECKER'
_mods_var = 'AMINO_TYPECHECK_MODS'
check_mods: List[str] = []
tc: Any = None

if _enable_var in os.environ:
    mods = os.environ.get(_mods_var) or 'amino'
    check_mods = mods.split(',')


class Loader(SourceFileLoader):

    def exec_module(self, module: ModuleType) -> None:
        from enforce import runtime_validation
        super().exec_module(module)
        if module.__package__ and module.__package__.split('.')[0] in check_mods:
            for clsname, cls in inspect.getmembers(module, inspect.isclass):
                for funname, fun in list((k, v) for k, v in cls.__dict__.items() if inspect.isfunction(v)):
                    try:
                        module.__dict__[funname] = runtime_validation(fun)
                    except Exception as e:
                        logging.debug(f'failed to apply runtime validation on {fun}: {e}')


def init_enforce() -> None:
    loader_details = (Loader, SOURCE_SUFFIXES)
    sys.path_hooks.insert(0, FileFinder.path_hook(loader_details))  # type: ignore


def init_typeguard() -> None:
    from typeguard import TypeChecker
    global tc
    tc = TypeChecker(check_mods)
    tc.start()


def boot() -> bool:
    if check_mods:
        if os.environ.get(_checker_var) == 'enforce':
            init_enforce()
        else:
            init_typeguard()
        return True
    else:
        return False

__all__ = ('boot', 'init_enforce', 'init_typeguard')



import abc
import re
from typing import TypeVar, Generic, Callable, Any

from amino.tc.base import TypeClass, tc_prop
from amino.func import ReplaceVal

F = TypeVar('F')
A = TypeVar('A')
B = TypeVar('B')


class Functor(Generic[F], TypeClass[F]):
    _map_re = re.compile('^map(\d+)$')

    @abc.abstractmethod
    def map(self, fa: F, f: Callable[[A], B]) -> F:
        ...

    def __truediv__(self, fa, f):
        return self.map(fa, f)

    def __getattr__(self, name: str) -> Any:
        match = self._map_re.match(name)
        if match is None:
            raise AttributeError(f'''`Functor` object has no attribute `{name}`''')
        return lambda *a: self.map_n(int(match.group(1)), *a)

    def map_n(self, num: int, fa: F, f: Callable[..., B]) -> F:
        def wrapper(args):
            if len(args) != num:
                msg = 'passed {} args to {}.map{}'
                name = self.__class__.__name__
                raise TypeError(msg.format(len(args), name, num))
            return f(*args)
        return self.map(fa, wrapper)

    def replace(self, fa: F, b: B) -> F:
        return self.map(fa, ReplaceVal(b))

    @tc_prop
    def void(self, fa: F) -> F:
        return self.replace(fa, None)

    def foreach(self, fa: F, f: Callable[[A], Any]) -> F:
        def effect(a: A) -> A:
            f(a)
            return a
        return self.map(fa, effect)

    __mod__ = foreach

__all__ = ('Functor',)
